"""
Vasicek Process
"""
from aleatory.processes.base import SPEulerMaruyama
import numpy as np
from scipy.stats import norm


class Vasicek(SPEulerMaruyama):
    r"""
    Vasicek Process

    .. image:: _static/vasicek_process_drawn.png


    A Vasicek process :math:`X = \{X : t \geq  0\}` is characterised by the following
    Stochastic Differential Equation

    .. math::

      dX_t = \theta(\mu - X_t) dt + \sigma X_t dW_t, \ \ \ \ \forall t\in (0,T],

    with initial condition :math:`X_0 = x_0`, where

    - :math:`\theta` is the speed of reversion
    - :math:`\mu` is the long term mean value.
    - :math:`\sigma>0` is the instantaneous volatility
    - :math:`W_t` is a standard Brownian Motion.


    Each :math:`X_t` follows a normal distribution.


    :param float theta: the parameter :math:`\theta` in the above SDE
    :param float mu: the parameter :math:`\mu` in the above SDE
    :param float sigma: the parameter :math:`\sigma>0` in the above SDE
    :param float initial: the initial condition :math:`x_0` in the above SDE
    :param float T: the right hand endpoint of the time interval :math:`[0,T]`
        for the process
    :param numpy.random.Generator rng: a custom random number generator
    """

    def __init__(self, theta=1.0, mu=3.0, sigma=0.5, initial=1.0, T=1.0, rng=None):
        super().__init__(T=T, rng=rng)
        self.theta = theta
        self.mu = mu
        self.sigma = sigma
        self.initial = initial
        self.n = 1.0
        self.dt = 1.0 * self.T / self.n
        self.times = None
        self.name = "Vasicek Process"

        def f(x, _):
            return self.theta * (self.mu - x)

        def g(x, _):
            return self.sigma

        self.f = f
        self.g = g

    def __str__(self):
        return "Vasicek process with parameters {speed}, {mean}, and {volatility} on [0, {T}].".format(
            T=str(self.T), speed=str(self.theta), mean=str(self.mu), volatility=str(self.sigma))

    def _process_expectation(self, times=None):
        if times is None:
            times = self.times
        return self.initial * np.exp((-1.0) * self.theta * times) + self.mu * (
                np.ones(len(times)) - np.exp((-1.0) * self.theta * times))

    def marginal_expectation(self, times=None):
        expectations = self._process_expectation(times=times)
        return expectations

    def _process_variance(self,times=None):
        if times is None:
            times = self.times
        variances = (self.sigma ** 2) * (1.0 / (2.0 * self.theta)) * (
                np.ones(len(times)) - np.exp(-2.0 * self.theta * times))
        return variances

    def marginal_variance(self, times=None):
        variances = self._process_variance(times=times)
        return variances

    def _process_stds(self):
        stds = np.sqrt(self.marginal_variance())
        return stds

    def process_stds(self):
        stds = self._process_stds()
        return stds

    def get_marginal(self, t):
        mu_x = self.initial * np.exp(-1.0 * self.theta * t) + self.mu * (1.0 - np.exp(-1.0 * self.theta * t))
        variance_x = (self.sigma ** 2) * (1.0 / (2.0 * self.theta)) * (1.0 - np.exp(-2.0 * self.theta * t))
        sigma_x = np.sqrt(variance_x)
        marginal = norm(loc=mu_x, scale=sigma_x)

        return marginal



from aleatory.processes import BrownianMotion, GBM, Vasicek, OUProcess, CIRProcess, CEVProcess
import numpy as np

SAVE = False
SAVE_PATH = '../docs/source/_static/'


def test_sample(n=100):
    process = BrownianMotion()
    sample = process.sample(n)
    times = process.times
    grid_times = np.linspace(0, process.T, n)
    assert len(times) == len(grid_times)
    assert len(times) == len(sample)
    for (t1, t2) in zip(times, grid_times):
        assert t1 == t2


def test_cir_convergence():
    process = CIRProcess()
    process.draw(n=100, N=200)
    process = CIRProcess(T=10)
    process.draw(n=100, N=200)


def test_rng():
    seed = 123
    bm1 = BrownianMotion(drift=2.0, scale=0.5, rng=np.random.default_rng(seed=seed))
    sim1 = bm1.simulate(n=100, N=1)

    bm2 = BrownianMotion(drift=2.0, scale=0.5, rng=np.random.default_rng(seed=seed))
    sim2 = bm2.simulate(n=100, N=1)

    for v1, v2 in zip(sim1[0], sim2[0]):
        assert v1 == v2


def test_figures_examples():
    bm = BrownianMotion()
    bmd = BrownianMotion(drift=-1.0, scale=0.5)
    gbm = GBM()
    vasicek = Vasicek()
    ouprocess = OUProcess()
    cirprocess = CIRProcess()
    cev = CEVProcess()

    for process in [bm, bmd, gbm, vasicek, ouprocess, cirprocess, cev]:

        process.plot(n=100, N=5, figsize=(9.5, 6), dpi=100)
        process.plot(n=100, N=5, title='My favourite figure', figsize=(9.5, 6), dpi=100)  # figure_with_title

        name = process.name.replace(" ", "_").lower()
        if SAVE:
            figure = process.plot(n=100, N=5, figsize=(9.5, 6), dpi=100)
            figure.savefig(SAVE_PATH + name + '_simple_plot.png')
            figure = process.draw(n=100, N=200, figsize=(12, 6), dpi=100)
            figure.savefig(SAVE_PATH + name + '_drawn.png')

        process.draw(n=100, N=200, figsize=(12, 6), dpi=100)
        process.draw(n=100, N=200, envelope=True, figsize=(12, 6), dpi=100)
        process.draw(n=100, N=200, marginal=False, figsize=(9.5, 6), dpi=100)
        process.draw(n=100, N=200, marginal=False, envelope=True, figsize=(9.5, 6), dpi=100)

        process.draw(n=100, N=200, figsize=(12, 6), dpi=100, title='My favourite figure')
        process.draw(n=100, N=200, envelope=True, figsize=(12, 6), dpi=100, title='My favourite figure')
        process.draw(n=100, N=200, marginal=False, figsize=(9.5, 6), dpi=100, title='My favourite figure')
        process.draw(n=100, N=200, marginal=False, envelope=True, figsize=(9.5, 6), dpi=100,
                     title='My favourite figure')


def test_quick_start():
    process = BrownianMotion()
    name = process.name.replace(" ", "_").lower()

    figure = process.plot(n=100, N=10, figsize=(9.5, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_01.png')

    figure = process.draw(n=100, N=200, figsize=(12, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_02.png')

    figure = process.draw(n=100, N=200, envelope=True, figsize=(12, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_03.png')

    figure = process.draw(n=100, N=200, marginal=False, figsize=(9.5, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_04.png', figsize=(12, 6), dpi=100)

    figure = process.draw(n=100, N=200, marginal=False, envelope=True, figsize=(9.5, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_05.png', figsize=(12, 6), dpi=100)

    figure = process.plot(n=100, N=200, style='ggplot', figsize=(12, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_06.png')

    figure = process.draw(n=100, N=200, style='Solarize_Light2', figsize=(12, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_07.png')

    figure = process.draw(n=100, N=200, colormap="cool", figsize=(12, 6), dpi=100)
    if SAVE:
        figure.savefig(SAVE_PATH + name + '_quickstart_08.png', dpi=300)



from numbers import Number
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import pandas as pd
import statsmodels.api as sm


def get_times(end, n):
    """Generate a linspace from 0 to end for n increments."""
    return np.linspace(0, end, n)


def check_positive_integer(n, name=""):
    """Ensure that the number is a positive integer."""
    if not isinstance(n, int):
        raise TypeError(f"{name} must be an integer.")
    if n <= 0:
        raise ValueError(f"{name} must be positive.")


def check_numeric(value, name=""):
    """Ensure that the value is numeric."""
    if not isinstance(value, Number):
        raise TypeError(f"{name} value must be a number.")


def check_positive_number(value, name=""):
    """Ensure that the value is a positive number."""
    check_numeric(value, name)
    if value <= 0:
        raise ValueError(f"{name} value must be positive.")


def check_increments(times):
    increments = np.diff(times)
    if np.any([t < 0 for t in times]):
        raise ValueError("Times must be nonnegative.")
    if np.any([t <= 0 for t in increments]):
        raise ValueError("Times must be strictly increasing.")
    return increments


def times_to_increments(times):
    """Ensure a positive, monotonically increasing sequence."""
    return check_increments(times)


def plot_paths(times, paths, style="seaborn-v0_8-whitegrid", title=None, **fig_kw):
    with plt.style.context(style):
        fig, ax = plt.subplots(**fig_kw)
        for p in paths:
            ax.plot(times, p)
        ax.set_title(title)
        ax.set_xlabel('$t$')
        ax.set_ylabel('$X(t)$')
        plt.show()
    return fig


def draw_paths(times, paths, N, expectations, title=None, KDE=False, marginal=False, marginalT=None, envelope=False,
               lower=None, upper=None, style="seaborn-v0_8-whitegrid", colormap="RdYlBu_r", **fig_kw):
    with plt.style.context(style):
        if marginal:
            fig = plt.figure(**fig_kw)
            gs = GridSpec(1, 5)
            ax1 = fig.add_subplot(gs[:4])
            ax2 = fig.add_subplot(gs[4:], sharey=ax1)

            last_points = [path[-1] for path in paths]
            cm = plt.colormaps[colormap]
            n_bins = int(np.sqrt(N))
            n, bins, patches = ax2.hist(last_points, n_bins, orientation='horizontal', density=True)
            bin_centers = 0.5 * (bins[:-1] + bins[1:])
            col = bin_centers - min(bin_centers)  # scale values to interval [0,1]
            col /= max(col)
            for c, p in zip(col, patches):
                plt.setp(p, 'facecolor', cm(c))
            my_bins = pd.cut(last_points, bins=bins, labels=range(len(bins) - 1), include_lowest=True)
            colors = [col[b] for b in my_bins]

            if KDE:
                kde = sm.nonparametric.KDEUnivariate(last_points)
                kde.fit()  # Estimate the densities
                ax2.plot(kde.density, kde.support, '--', lw=1.75, alpha=0.6, label='$X_T$  KDE', zorder=10)
                ax2.axhline(y=np.mean(last_points), linestyle='--', lw=1.75, label=r'$\overline{X_T}$')
            else:
                marginaldist = marginalT
                x = np.linspace(marginaldist.ppf(0.005), marginaldist.ppf(0.995), 100)
                ax2.plot(marginaldist.pdf(x), x, '-', lw=1.75, alpha=0.6, label='$X_T$ pdf')
                ax2.axhline(y=marginaldist.mean(), linestyle='--', lw=1.75, label='$E[X_T]$')

            plt.setp(ax2.get_yticklabels(), visible=False)
            ax2.set_title('$X_T$')
            ax2.legend()

            for i in range(N):
                ax1.plot(times, paths[i], '-', lw=1.0, color=cm(colors[i]))
            ax1.plot(times, expectations, '--', lw=1.75, label='$E[X_t]$')
            if envelope:
                ax1.fill_between(times, upper, lower, alpha=0.25, color='grey')
            plt.subplots_adjust(wspace=0.025, hspace=0.025)

        else:
            fig, ax1 = plt.subplots(**fig_kw)
            for i in range(N):
                ax1.plot(times, paths[i], '-', lw=1.0)
            ax1.plot(times, expectations, '--', lw=1.75, label='$E[X_t]$')
            if envelope:
                ax1.fill_between(times, upper, lower, color='grey', alpha=0.25)

        fig.suptitle(title)
        ax1.set_title('Simulated Paths $X_t, t \in [t_0, T]$')  # Title
        ax1.set_xlabel('t')
        ax1.set_ylabel('X(t)')
        ax1.legend()
        plt.show()

    return fig



from aleatory.processes import BrownianMotion
import numpy as np
import pytest


class TestMarginalFunctions:

    def test_marginals_BM(self,drif=1.0, scale=2.0):
        grid_times = np.linspace(0, 1.0, 100)
        process = BrownianMotion(drift=drif, scale=scale)
        means = process.marginal_expectation(grid_times)
        variances = process.marginal_variance(grid_times)

        for m, t in zip(means, grid_times):
            assert m == drif*t

        for v, t in zip(variances, grid_times):
            assert v == (scale ** 2) * t

        process.draw(n=100, N=50)
        means = process.marginal_expectation()
        for m, t in zip(means, process.times):
            assert m == drif*t



from aleatory.processes.euler_maruyama.vasicek import Vasicek
from aleatory.processes.euler_maruyama.ornstein_uhlenbeck import OUProcess
from aleatory.processes.euler_maruyama.cir_process import CIRProcess
from aleatory.processes.euler_maruyama.cev_process import CEVProcess
from aleatory.processes.analytical.geometric_brownian import GBM
from aleatory.processes.analytical.brownian_motion import BrownianMotion


"""
Constant Elasticity Variance Process
"""
from aleatory.processes.base import SPEulerMaruyama
import numpy as np
from aleatory.utils.utils import draw_paths


class CEVProcess(SPEulerMaruyama):
    r"""
    CEV or constant elasticity of variance process

    .. image:: _static/cev_process_drawn.png


    A CEV process  :math:`X = \{X : t \geq  0\}` is characterised by the following
    Stochastic Differential Equation

    .. math::

      dX_t = \mu X_t dt + \sigma X_t^{\gamma} dW_t, \ \ \ \ \forall t\in (0,T],

    with initial condition :math:`X_0 = x_0`, where

    - :math:`\mu` is the drift
    - :math:`\sigma>0` is the scale of the volatility
    - :math:`\gamma\geq 0` is the elasticity term
    - :math:`W_t` is a standard Brownian Motion.


    :param float mu: the parameter :math:`\mu` in the above SDE
    :param float sigma: the parameter :math:`\sigma>0` in the above SDE
    :param float gamma: the parameter :math:`\gamma` in the above SDE
    :param float initial: the initial condition :math:`x_0` in the above SDE
    :param float T: the right hand endpoint of the time interval :math:`[0,T]`
        for the process
    :param numpy.random.Generator rng: a custom random number generator
    """

    def __init__(self, gamma=1.5, mu=0.5, sigma=0.1, initial=1.0, T=1.0, rng=None):
        super().__init__(T=T, rng=rng)
        self.gamma = gamma
        self.mu = mu
        self.sigma = sigma
        self.initial = initial
        self.n = 1.0
        self.dt = 1.0 * self.T / self.n
        self.times = np.arange(0.0, self.T + self.dt, self.dt)
        self.name = "CEV Process"
        self.paths = None
        self._marginals = None

        def f(x, _):
            return self.mu - 0.5 * (self.sigma ** 2) * np.exp(2.0 * (self.gamma - 1.0) * x)
            # return self.mu * x

        def g(x, _):
            return self.sigma * np.exp((self.gamma - 1.0) * x)
            # return self.sigma * (x ** self.gamma)

        self.f = f
        self.g = g

    @property
    def sigma(self):
        return self._sigma

    @sigma.setter
    def sigma(self, value):
        if value < 0:
            raise ValueError("sigma cannot be negative")
        self._sigma = value

    @property
    def gamma(self):
        return self._gamma

    @gamma.setter
    def gamma(self, value):
        if value < 0:
            raise ValueError("gamma cannot be negative")
        self._gamma = value

    def __str__(self):
        return "CEV process with parameters {gamma}, {drift}, and {volatility} on [0, {T}].".format(
            T=str(self.T), gamma=str(self.gamma), drift=str(self.mu), volatility=str(self.sigma))

    def _get_empirical_marginal_samples(self):
        if self.paths is None:
            self.simulate(self.n, self.N)

        empirical_marginal_samples = np.array(self.paths).transpose()
        return empirical_marginal_samples

    def get_marginal(self, t):
        pass

    def estimate_expectations(self):
        if self._marginals is None:
            self._marginals = self._get_empirical_marginal_samples()

        empirical_means = [np.mean(m) for m in self._marginals]
        return empirical_means

    def estimate_variances(self):
        if self._marginals is None:
            self._marginals = self._get_empirical_marginal_samples()
        empirical_vars = [np.var(m) for m in self._marginals]
        return empirical_vars

    def estimate_stds(self):
        variances = self.estimate_variances()
        stds = [np.sqrt(var) for var in variances]
        return stds

    def estimate_quantiles(self, q):
        if self._marginals is None:
            self._marginals = self._get_empirical_marginal_samples()
        empirical_quantiles = [np.quantile(m, q) for m in self._marginals]
        return empirical_quantiles

    def _process_expectation(self):
        return self.estimate_expectations()

    def process_expectation(self):
        return self._process_expectation()

    def _process_variance(self):
        return self.estimate_variances()

    def process_variance(self):
        return self._process_variance()

    def _process_stds(self):
        stds = np.sqrt(self.process_variance())
        return stds

    def process_stds(self):
        stds = self._process_stds()
        return stds

    def draw(self, n, N, marginal=True, envelope=False, title=None, **fig_kw):
        self.simulate(n, N)
        expectations = self.estimate_expectations()

        if envelope:
            lower = self.estimate_quantiles(0.005)
            upper = self.estimate_quantiles(0.995)
        else:
            lower = None
            upper = None

        chart_title = title if title else self.name

        if marginal:
            fig = draw_paths(times=self.times, paths=self.paths, N=N, title=chart_title, KDE=True, marginal=marginal,
                             expectations=expectations, envelope=envelope, lower=lower, upper=upper,
                             **fig_kw)
        else:
            fig = draw_paths(times=self.times, paths=self.paths, N=N, title=chart_title,
                             expectations=expectations, marginal=marginal, **fig_kw)

        return fig

    def sample(self, n):
        return self._sample_em_process(n, log=True)



"""Brownian Motion"""
import numpy as np
from scipy.stats import norm

from aleatory.processes.base import SPExplicit
from aleatory.processes.analytical.gaussian import GaussianIncrements
from aleatory.utils.utils import check_positive_number, check_numeric, get_times


class BrownianMotion(SPExplicit):
    r"""Brownian motion

    .. image:: _static/brownian_motion_drawn.png


    A standard Brownian motion    :math:`\{W_t : t \geq 0\}` is defined by the following properties:

    1. Starts at zero, i.e. :math:`W(0) = 0`
    2. Independent increments
    3. :math:`W(t) - W(s)` follows a Gaussian distribution :math:`N(0, t-s)`
    4. Almost surely continuous

    A more general version of a Brownian motion, is the Drifted Brownian Motion which  is defined
    by the following SDE

    .. math::

        dX_t = \mu dt + \sigma dW_t \ \ \ \   t\in (0,T]


    with initial condition :math:`X_0 = x_0\in\mathbb{R}`, where

    - :math:`\mu` is the drift
    - :math:`\sigma>0` is the volatility
    - :math:`W_t` is a standard Brownian Motion.

    Clearly, the solution to this equation can be written as

    .. math::

        X_t = x_0 +  \mu t + \sigma W_t \ \ \ \ t \in [0,T]

    and each :math:`X_t \sim N(\mu t, \sigma^2 t)`.

    :param float drift: the parameter :math:`\mu` in the above SDE
    :param float scale: the parameter :math:`\sigma>0` in the above SDE
    :param float initial: the initial condition :math:`x_0` in the above SDE
    :param float T: the right hand endpoint of the time interval :math:`[0,T]`
        for the process
    :param numpy.random.Generator rng: a custom random number generator

    """

    def __init__(self, drift=0.0, scale=1.0, T=1.0, rng=None):
        super().__init__(T=T, rng=rng, initial=0.0)
        self.drift = drift
        self.scale = scale
        self.name = "Brownian Motion" if drift == 0.0 else "Brownian Motion with Drift"
        self.n = None
        self.times = None
        self.gaussian_increments = GaussianIncrements(T=self.T, rng=self.rng)

    @property
    def drift(self):
        return self._drift

    @drift.setter
    def drift(self, value):
        check_numeric(value, "Drift")
        self._drift = value

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        check_positive_number(value, "Scale")
        self._scale = value

    def _sample_brownian_motion(self, n):
        self.n = n
        self.times = get_times(self.T, self.n)
        bm = np.cumsum(self.scale * self.gaussian_increments.sample(n - 1))
        bm = np.insert(bm, 0, [0])
        if self.drift == 0:
            return bm
        else:
            return self.times * self.drift + bm

    def sample(self, n):
        """
        Generates a discrete time sample from a Brownian Motion instance.

        :param n: the number of steps
        :return: numpy array
        """
        return self._sample_brownian_motion(n)

    def _sample_brownian_motion_at(self, times):
        self.times = times
        bm = np.cumsum(self.scale * self.gaussian_increments.sample_at(times))

        if times[0] != 0:
            bm = np.insert(bm, 0, [0])

        if self.drift != 0:
            bm += [self.drift * t for t in times]

        return bm

    def sample_at(self, times):
        """
        Generates a sample from a Brownian motion at the specified times.

        :param times: the times which define the sample
        :return: numpy array
        """
        return self._sample_brownian_motion_at(times)

    def _process_expectation(self, times=None):
        if times is None:
            times = self.times
        return self.drift * times

    def marginal_expectation(self, times=None):
        expectations = self._process_expectation(times=times)
        return expectations

    def _process_variance(self, times=None):
        if times is None:
            times = self.times
        return (self.scale ** 2) * times

    def marginal_variance(self, times):
        variances = self._process_variance(times=times)
        return variances

    def _process_stds(self):
        return self.scale * np.sqrt(self.times)

    def process_stds(self):
        stds = self._process_stds()
        return stds

    def get_marginal(self, t):
        marginal = norm(loc=self.drift * t, scale=self.scale * np.sqrt(t))
        return marginal

    def draw(self, n, N, marginal=True, envelope=False, type='3sigma', title=None, **fig_kw):
        """
        Simulates and plots paths/trajectories from the instanced stochastic process.

        Produces different kind of visualisation illustrating the following elements:

        - times versus process values as lines
        - the expectation of the process across time
        - histogram showing the empirical marginal distribution :math:`X_T` (optional when ``marginal = True``)
        - probability density function of the marginal distribution :math:`X_T` (optional when ``marginal = True``)
        - envelope of confidence intervals across time (optional when ``envelope = True``)


        :param n: number of steps in each path
        :param N: number of paths to simulate
        :param marginal: bool, default: True
        :param envelope: bool, default: False
        :param type: string, default: '3sigma'
        :param title: string to customise plot title
        :return:
        """

        if type == '3sigma':
            return self._draw_3sigmastyle(n=n, N=N, marginal=marginal, envelope=envelope, title=title, **fig_kw)
        elif type == 'qq':
            return self._draw_qqstyle(n, N, marginal=marginal, envelope=envelope, title=title, **fig_kw)
        else:
            raise ValueError



from aleatory.processes.base import SPEulerMaruyama
import numpy as np
from scipy.stats import ncx2


class CIRProcess(SPEulerMaruyama):
    r"""
    Cox–Ingersoll–Ross Process

    .. image:: _static/cir_process_drawn.png

    A Cox–Ingersoll–Ross process :math:`X = \{X : t \geq  0\}` is characterised by the following
    Stochastic Differential Equation

    .. math::

      dX_t = \theta(\mu - X_t) dt + \sigma \sqrt{X_t} dW_t, \ \ \ \ \forall t\in (0,T],

    with initial condition :math:`X_0 = x_0`, where

    - :math:`\theta` is the rate of mean reversion
    - :math:`\mu` is the long term mean value.
    - :math:`\sigma>0` is the instantaneous volatility
    - :math:`W_t` is a standard Brownian Motion.


    Each :math:`X_t` follows a non-central chi-square distribution.


    :param float theta: the parameter :math:`\theta` in the above SDE
    :param float mu: the parameter :math:`\mu` in the above SDE
    :param float sigma: the parameter :math:`\sigma>0` in the above SDE
    :param float initial: the initial condition :math:`x_0` in the above SDE
    :param float T: the right hand endpoint of the time interval :math:`[0,T]`
        for the process
    :param numpy.random.Generator rng: a custom random number generator

    """

    def __init__(self, theta=1.0, mu=2.0, sigma=0.5, initial=5.0, T=1.0, rng=None):
        super().__init__(T=T, rng=rng, initial=initial)
        self.theta = theta
        self.mu = mu
        self.sigma = sigma
        self.n = 1.0
        self.dt = 1.0 * self.T / self.n
        self.times = np.arange(0.0, self.T + self.dt, self.dt)
        self.name = "CIR Process"

        def f(x, _):
            # return self.theta * (self.mu - x)
            return np.exp(-x) * (self.theta * (self.mu - np.exp(x)) - 0.5 * self.sigma ** 2)

        def g(x, _):
            # return self.sigma * np.sqrt(x)
            return self.sigma * np.exp(-0.5 * x)

        self.f = f
        self.g = g

    @property
    def theta(self):
        return self._theta

    @theta.setter
    def theta(self, value):
        if value < 0:
            raise ValueError("theta must be positive")
        self._theta = value

    @property
    def sigma(self):
        return self._sigma

    @sigma.setter
    def sigma(self, value):
        if value <= 0:
            raise ValueError("sigma has to be positive")
        if 2 * self.theta * self.mu <= value ** 2:
            raise ValueError("Condition 2*theta*mu >= sigma**2 must be satisfied")
        self._sigma = value

    def __str__(self):
        return "Cox–Ingersoll–Ross process with parameters {speed}, {mean}, and {volatility} on [0, {T}].".format(
            T=str(self.T), speed=str(self.theta), mean=str(self.mu), volatility=str(self.sigma))

    def _process_expectation(self, times=None):
        if times is None:
            times = self.times
        expectations = self.initial * np.exp((-1.0) * self.theta * times) + self.mu * (
                np.ones(len(times)) - np.exp((-1.0) * self.theta * times))
        return expectations

    def marginal_expectation(self, times=None):
        expectations = self._process_expectation(times=times)
        return expectations

    def _process_variance(self, times=None):
        if times is None:
            times = self.times
        variances = (self.initial * self.sigma ** 2 / self.theta) * (
                np.exp(-1.0 * self.theta * times) - np.exp(-2.0 * self.theta * times)) + (
                            self.mu * self.sigma ** 2 / 2 * self.theta) * (
                            (np.ones(len(times)) - np.exp(-1.0 * self.theta * times)) ** 2)
        return variances

    def marginal_variance(self, times=None):
        variances = self._process_variance(times=times)
        return variances

    def _process_stds(self):
        stds = np.sqrt(self._process_variance())
        return stds

    def process_stds(self):
        stds = self._process_stds()
        return stds

    def get_marginal(self, t):
        a = self.theta
        b = self.mu
        sigma = self.sigma

        c = 2.0 * a / ((1.0 - np.exp(-1.0 * a * t)) * sigma ** 2)
        df = 4.0 * a * b / sigma ** 2
        nc = 2.0 * c * self.initial * np.exp(-1.0 * a * t)
        scale = 1.0 / (2 * c)
        marginal = ncx2(df, nc, scale=scale)

        return marginal

    def sample(self, n):
        return self._sample_em_process(n, log=True)



from abc import ABC
from abc import abstractmethod

import numpy as np

from aleatory.utils.utils import check_positive_number
from aleatory.utils.utils import plot_paths, draw_paths, check_positive_integer, get_times


class BaseProcess(ABC):
    def __init__(self, rng=None):
        self.rng = rng

    @property
    def rng(self):
        if self._rng is None:
            return np.random.default_rng()
        return self._rng

    @rng.setter
    def rng(self, value):
        if value is None:
            self._rng = None
        elif isinstance(value, (np.random.RandomState, np.random.Generator)):
            self._rng = value
        else:
            raise TypeError("rng must be of type `numpy.random.Generator`")


class StochasticProcess(BaseProcess, ABC):
    """
    Base class for all one-factor stochastic processes classes.
    All processes of this type are defined on a finite interval $[0,T]$.
    """

    def __init__(self, T=1.0, rng=None):
        super().__init__(rng=rng)
        self.T = T

    @property
    def T(self):
        """End time of the process."""
        return self._T

    @T.setter
    def T(self, value):
        check_positive_number(value, "Time end")
        self._T = float(value)


class SPExplicit(StochasticProcess):

    def __init__(self, initial=0.0, name=None, T=1.0, rng=None):
        super().__init__(T=T, rng=rng)
        self.initial = initial
        self.name = name
        self.n = None
        self.times = None
        self.N = None
        self.paths = None

    @abstractmethod
    def sample(self, n):  # pragma: no cover
        pass

    def simulate(self, n, N):
        """
        Simulate paths/trajectories from the instanced stochastic process.

        :param n: number of steps in each path
        :param N: number of paths to simulate
        :return: list with N paths (each one is a np.array of size n)
        """
        self.n = n
        self.N = N
        self.paths = [self.sample(n) for _ in range(N)]
        return self.paths

    @abstractmethod
    def get_marginal(self, t):
        pass

    @abstractmethod
    def _process_expectation(self):
        pass

    @abstractmethod
    def _process_variance(self):
        pass

    @abstractmethod
    def _process_stds(self):
        pass

    def plot(self, n, N, title=None, **fig_kw):
        """
        Simulates and plots paths/trajectories from the instanced stochastic process.
        Simple plot of times versus process values as lines and/or markers.

        :param n: number of steps in each path
        :param N: number of paths to simulate
        :param title: string to customise plot title
        :return:
        """
        self.simulate(n, N)
        if title:
            figure = plot_paths(self.times, self.paths, title=title, **fig_kw)
        else:
            figure = plot_paths(self.times, self.paths, title=self.name, **fig_kw)
        return figure

    def _draw_paths(self, n, N, marginal=False, envelope=False, type=None, title=None, **fig_kw):
        self.simulate(n, N)
        expectations = self._process_expectation()

        if envelope:
            if type == '3sigma':
                stds = self._process_stds()
                upper = expectations + 3.0 * stds
                lower = expectations - 3.0 * stds
            else:
                marginals = [self.get_marginal(t) for t in self.times[1:]]
                upper = [self.initial] + [m.ppf(0.005) for m in marginals]
                lower = [self.initial] + [m.ppf(0.995) for m in marginals]
        else:
            upper = None
            lower = None

        if marginal:
            marginalT = self.get_marginal(self.T)
        else:
            marginalT = None

        chart_title = title if title else self.name
        fig = draw_paths(times=self.times, paths=self.paths, N=N, title=chart_title, expectations=expectations,
                         marginal=marginal, marginalT=marginalT, envelope=envelope, lower=lower, upper=upper,
                         **fig_kw)
        return fig

    def _draw_qqstyle(self, n, N, marginal=False, envelope=False, title=None,
                      **fig_kw):

        fig = self._draw_paths(n=n, N=N, marginal=marginal, envelope=envelope, type='qq', title=title, **fig_kw)
        return fig

    def _draw_3sigmastyle(self, n, N, marginal=False, envelope=False, title=None, **fig_kw):

        fig = self._draw_paths(n=n, N=N, marginal=marginal, envelope=envelope, type='3sigma', title=title, **fig_kw)
        return fig

    def draw(self, n, N, marginal=True, envelope=False, title=None, **fig_kw):
        """
        Simulates and plots paths/trajectories from the instanced stochastic process.
        Visualisation shows
        - times versus process values as lines
        - the expectation of the process across time
        - histogram showing the empirical marginal distribution :math:`X_T`
        - probability density function of the marginal distribution :math:`X_T`
        - envelope of confidence intervals

        :param n: number of steps in each path
        :param N: number of paths to simulate
        :param marginal: bool, default: True
        :param envelope: bool, default: False
        :return:
        """
        return self._draw_qqstyle(n, N, marginal=marginal, envelope=envelope, title=title, **fig_kw)


class SPEulerMaruyama(SPExplicit):
    def __int__(self, f=None, g=None, initial=0.0, T=1.0, rng=None):
        super().__init__(T=T, rng=rng, initial=initial)
        self.f = f
        self.g = g
        self.n = None
        self.dt = None
        self.times = None
        self.name = None

    def _sample_em_process(self, n, log=False):
        check_positive_integer(n)
        self.n = n
        self.dt = 1.0 * self.T / self.n
        self.times = get_times(self.T, self.n)
        dws = self.rng.normal(scale=np.sqrt(self.dt), size=self.n)

        if log:
            origin = np.log(self.initial)
        else:
            origin = self.initial

        path = [origin]
        previous = origin

        for (t, dw) in zip(self.times[1:], dws[1:]):
            previous += self.f(previous, t) * self.dt + self.g(previous, t) * dw
            path.append(previous)

        if log:
            return np.exp(path)

        return path

    def sample(self, n):
        return self._sample_em_process(n)



"""
Ornstein-Uhlenbeck Process
"""
from aleatory.processes.euler_maruyama.vasicek import Vasicek


class OUProcess(Vasicek):
    r"""
    Ornstein–Uhlenbeck Process

    .. image:: _static/ornstein–uhlenbeck_process_drawn.png

    An Ornstein–Uhlenbeck process :math:`X = \{X : t \geq  0\}` is characterised by the following
    Stochastic Differential Equation

    .. math::

      dX_t = -\theta X_t dt + \sigma X_t dW_t, \ \ \ \ \forall t\in (0,T],

    with initial condition :math:`X_0 = x_0`, where

    - :math:`\theta` is the speed of reversion
    - :math:`\sigma>0` is the instantaneous volatility
    - :math:`W_t` is a standard Brownian Motion.


    Each :math:`X_t` follows a normal distribution.


    :param float theta: the parameter :math:`\theta` in the above SDE
    :param float sigma: the parameter :math:`\sigma>0` in the above SDE
    :param float initial: the initial condition :math:`x_0` in the above SDE
    :param float T: the right hand endpoint of the time interval :math:`[0,T]`
        for the process
    :param numpy.random.Generator rng: a custom random number generator

    """

    def __init__(self, theta=1.0, sigma=0.5, initial=1.0, T=1.0, rng=None):
        super().__init__(theta=theta, mu=0.0, sigma=sigma, initial=initial, T=T, rng=rng)
        self.n = 1.0
        self.dt = 1.0 * self.T / self.n
        self.times = None
        self.name = "Ornstein–Uhlenbeck process"

    def __str__(self):
        return "Ornstein–Uhlenbeck process with parameters {speed}, and {volatility} on [0, {T}].".format(
            T=str(self.T), speed=str(self.theta), volatility=str(self.sigma))


import numpy as np

from aleatory.processes.base import StochasticProcess
from aleatory.utils.utils import check_positive_integer, times_to_increments, get_times


class GaussianIncrements(StochasticProcess):
    """
    Gaussian increments
    """
    def __int__(self, T=1.0, rng=None):
        super().__init__(T=T, rng=rng)
        self.name = "Gaussian Noise"
        self.times = None
        self.n = None
        self.paths = None

    def __str__(self):
        return "Gaussian noise process on interval [0, {T}]".format(T=str(self.T))

    def __repr__(self):
        return "GaussianNoise(T={T})".format(T=str(self.T))

    def _sample_gaussian_noise(self, n):
        """
        Generates a random sample of size n from N(0, T/n)
        :param n: number of increments
        :return:
        """
        check_positive_integer(n)
        self.n = n
        delta_t = 1.0 * self.T / self.n
        self.times = get_times(self.T, self.n)
        noise = self.rng.normal(scale=np.sqrt(delta_t), size=self.n)

        return noise

    def _sample_gaussian_noise_at(self, times):
        """
        Generates random Gaussian increments corresponding to the specified times.
        :param times: given times
        :return:
        """
        if times[0] != 0:
            times = np.concatenate(([0], times))
        increments = times_to_increments(times)
        self.times = times
        noise = np.array([self.rng.normal(scale=np.sqrt(inc)) for inc in increments])
        noise = np.concatenate(([0], noise))

        return noise

    def sample(self, n):
        """
        Generates a random sample of size n from N(0, T/n)
        :param n:
        :return:
        """
        return self._sample_gaussian_noise(n)

    def sample_at(self, times):
        """
        Generates random Gaussian increments corresponding to the specified times
        $t_1, \cdots, t_m$
        Note that the first increment always starts with zero.
        :param times:
        :return:
        """
        return self._sample_gaussian_noise_at(times)



"""
Geometric Brownian Motion
"""
import numpy as np
from scipy.stats import lognorm

from aleatory.processes.base import SPExplicit
from aleatory.processes.analytical.brownian_motion import BrownianMotion
from aleatory.utils.utils import check_positive_number, check_numeric, get_times, check_positive_integer


class GBM(SPExplicit):
    r"""Geometric Brownian Motion

    .. image:: _static/geometric_brownian_motion_drawn.png


    A Geometric Brownian Motion :math:`\{X(t) : t \geq  0\}` is characterised by
    the following SDE.

    .. math::
        dX_t = \mu X_t dt + \sigma X_t dW_t \ \ \ \ t\in (0,T]


    with initial condition :math:`X_0 = x_0\geq0`, where

    - :math:`\mu` is the drift
    - :math:`\sigma>0` is the volatility
    - :math:`W_t` is a standard Brownian Motion.


    The solution to this equation can be written as

    .. math::

        X_t = x_0\exp\left((\mu + \frac{\sigma^2}{2} )t +\sigma W_t\right)




    and each :math:`X_t` follows a log-normal distribution.

    :param float drift: the parameter :math:`\mu` in the above SDE
    :param float volatility: the parameter :math:`\sigma>0` in the above SDE
    :param float initial: the initial condition :math:`x_0` in the above SDE
    :param float T: the right hand endpoint of the time interval :math:`[0,T]`
        for the process
    :param numpy.random.Generator rng: a custom random number generator

    """

    def __init__(self, drift=1.0, volatility=0.5, initial=1.0, T=1.0, rng=None):
        super().__init__(T=T, rng=rng, initial=initial)
        self.drift = drift
        self.volatility = volatility
        self._brownian_motion = BrownianMotion(T=T, rng=rng)
        self.name = "Geometric Brownian Motion"
        self.n = None
        self.times = None

    def __str__(self):
        return "Geometric Brownian motion with drift {d} and volatility {v} on [0, {T}].".format(
            T=str(self.T), d=str(self.drift), v=str(self.volatility))

    def __repr__(self):
        return "GeometricBrownianMotion(drift={d}, volatility={v}, T={T})".format(
            T=str(self.T), d=str(self.drift), v=str(self.volatility))

    @property
    def drift(self):
        """Geometric Brownian motion drift parameter."""
        return self._drift

    @drift.setter
    def drift(self, value):
        check_numeric(value, "Drift")
        self._drift = value

    @property
    def volatility(self):
        """Geometric Brownian motion volatility parameter."""
        return self._volatility

    @volatility.setter
    def volatility(self, value):
        check_positive_number(value, "Volatility")
        self._volatility = value

    @property
    def initial(self):
        """Geometric Brownian motion initial point."""
        return self._initial

    @initial.setter
    def initial(self, value):
        check_positive_number(value, "Initial Point")
        self._initial = value

    def _sample_geometric_brownian_motion(self, n):
        """Generate a realization of a geometric Brownian motion."""
        check_positive_integer(n)
        check_positive_number(self.initial, "Initial")
        self.n = n
        self.times = get_times(self.T, n)
        return self.initial * np.exp((self.drift - 0.5 * self.volatility ** 2) * self.times
                                     + self.volatility * self._brownian_motion.sample(n))

    def _sample_geometric_brownian_motion_at(self, times):
        """Generate a realization of a Geometric Brownian motion."""
        self.times = times
        return self.initial * np.exp((self.drift - 0.5 * self.volatility ** 2) * times
                                     + self.volatility * self._brownian_motion.sample_at(times))

    def sample(self, n):
        """Generate a realization.
        """
        return self._sample_geometric_brownian_motion(n)

    def sample_at(self, times):
        """Generate a realization using specified times.
        """
        return self._sample_geometric_brownian_motion_at(times)

    def _process_expectation(self, times=None):
        if times is None:
            times = self.times
        return self.initial * np.exp(self.drift * times)

    def marginal_expectation(self, times=None):
        expectations = self._process_expectation(times=times)
        return expectations

    def _process_variance(self, times=None):
        if times is None:
            times = self.times
        variances = (self.initial ** 2) * np.exp(2 * self.drift * times) * (
                np.exp(times * self.volatility ** 2) - 1)
        return variances

    def marginal_variance(self, times=None):
        variances = self._process_variance(times=times)
        return variances

    def _process_stds(self):
        variances = self.marginal_variance()
        stds = np.sqrt(variances)
        return stds

    def get_marginal(self, t):
        mu_x = np.log(self.initial) + (self.drift - 0.5 * self.volatility ** 2) * t
        sigma_x = self.volatility * np.sqrt(t)
        marginal = lognorm(s=sigma_x, scale=np.exp(mu_x))

        return marginal



from aleatory.processes import CIRProcess, CEVProcess


def test_transformation_em():
    cir = CIRProcess(theta=1.0, mu=2.0, sigma=0.25)
    cev = CEVProcess(gamma=1.5, mu=0.50, sigma=0.1, initial=1.0)

    cir.draw(n=100, N=200)
    cev.draw(n=100, N=200)
